import gzip
import sys
import enum


import symspellpy
from symspellpy import SymSpell, Verbosity

from tqdm import tqdm

from ordered_set import OrderedSet


class SymSpellMatchResultType(enum.Enum):
    SUCCESS = enum.auto()
    SUCCESS_BARCODE_ALREADY_CORRECT = enum.auto()
    AMBIGUOUS = enum.auto()
    NO_MATCH_IN_LEV_DIST = enum.auto()


def _load_barcode_allowlist_helper(open_file_handle, barcode_set, disable_pbar=None):
    if disable_pbar is None:
        disable_pbar = not sys.stdin.isatty()
    for line in tqdm(open_file_handle,
                     desc="Loading barcode allow list", unit=" barcode", colour="green", file=sys.stderr, leave=True,
                     disable=disable_pbar):
        bc = line.decode("utf-8").strip()
        barcode_set.add(bc)


def load_barcode_allowlist(filename, disable_pbar=None):
    barcodes = OrderedSet()

    if filename is not None:
        if filename.endswith(".gz"):
            with gzip.open(filename, 'rb') as f:
                _load_barcode_allowlist_helper(f, barcodes, disable_pbar)
        else:
            with open(filename, 'rb') as f:
                _load_barcode_allowlist_helper(f, barcodes, disable_pbar)

    return barcodes


def generate_symspell_index(wl_fname, max_dist_allowed, prefix_len=16, freq_list_sep="\t"):
    # Code generously donated from Victoria Popic
    sym_spell = SymSpell(max_dictionary_edit_distance=max_dist_allowed, prefix_length=prefix_len)
    sym_spell._distance_algorithm = symspellpy.editdistance.DistanceAlgorithm.LEVENSHTEIN_FAST
    sym_spell.load_dictionary(wl_fname, 0, 1, separator=freq_list_sep)  # positions of the term and count, respectively
    return sym_spell


def find_match_symspell(barcode, barcode_allow_list, sym_spell_index, dist_thr):
    # Adapted from code generously donated from Victoria Popic
    # using faster set lookups to find exact matches
    # for the most common case
    if barcode in barcode_allow_list:  # there is an exact match
        return barcode, 0, SymSpellMatchResultType.SUCCESS_BARCODE_ALREADY_CORRECT

    matches = sym_spell_index.lookup(barcode, Verbosity.CLOSEST, max_edit_distance=dist_thr)
    if not matches:
        return None, None, SymSpellMatchResultType.NO_MATCH_IN_LEV_DIST
    elif len(matches) > 1:
        return None, None, SymSpellMatchResultType.AMBIGUOUS

    return matches[0].term, matches[0].distance, SymSpellMatchResultType.SUCCESS



import matplotlib
import matplotlib.pyplot as plt

import numpy as np

# Make big figures:
gFIG_SIZE_in = [14, 10]

# Set plotting defaults:
gPLOT_PARAMS = {
    "legend.fontsize": "x-large",
    "figure.figsize": gFIG_SIZE_in,
    "axes.labelsize": "x-large",
    "axes.titlesize": "x-large",
    "xtick.labelsize": "x-large",
    "ytick.labelsize": "x-large"
}
matplotlib.rcParams.update(gPLOT_PARAMS)
plt.rcParams.update(gPLOT_PARAMS)

# Some single-place definitions of sizes for plots / figures:
gFONT_SIZE_UNITS = "pt"
gTITLE_FONT_SIZE = 36
gAXIS_LABEL_FONT_SIZE = 24
gTICK_LABEL_FONT_SIZE = 16
gTEXT_FONT_SIZE = 16

# To track global figure number creation:
gFIG_NUM = 0


def fix_plot_visuals(fig,
                     titlesize=gTITLE_FONT_SIZE,
                     labelsize=gAXIS_LABEL_FONT_SIZE,
                     ticklabelsize=gTICK_LABEL_FONT_SIZE,
                     textsize=gTEXT_FONT_SIZE,
                     tight_rect=None):
    """Fix the plot elements to be appropriate sizes for a slide / presentation."""

    if not textsize:
        textsize = ticklabelsize

    for ax in fig.get_axes():

        for ticklabel in (ax.get_xticklabels()):
            ticklabel.set_fontsize(ticklabelsize)
        for ticklabel in (ax.get_yticklabels()):
            ticklabel.set_fontsize(ticklabelsize)
        for c in ax.get_children():
            if c.__class__ == matplotlib.text.Text:
                c.set_fontsize(textsize)

        ax.xaxis.get_label().set_fontsize(labelsize)
        ax.yaxis.get_label().set_fontsize(labelsize)
        ax.title.set_fontsize(titlesize)

    for c in fig.get_children():
        if c.__class__ == matplotlib.legend.Legend:
            c.prop.set_size(ticklabelsize)
            c.get_title().set_size(ticklabelsize)

    if tight_rect:
        fig.tight_layout(rect=tight_rect)
    else:
        fig.tight_layout()


def save_figure(fig=None, name=None, prefix=None, suffix=None, fig_dir=None):
    """Save the current figure in the pyplot buffer to disk as both svg and png."""
    global gFIG_NUM

    if not fig_dir:
        fig_dir = ""
    elif not fig_dir.endswith("/"):
        fig_dir = f"{fig_dir}/"

    if not fig:
        fig = plt.gcf()

    if not name:
        # Get our name from the figure we're given:
        name = [ax.get_title() for ax in fig.axes if ax.get_title() and len(ax.get_title()) > 0][0]

    # Sanitize our name for writing to disk:
    name = name.replace(" ", "_").replace("\n", "_").replace("\t", "_").replace("=", "-").replace("(", "").replace(")", "")

    # Add prefix and suffix to the name if we have them defined:
    if prefix:
        name = f"{prefix}_{gFIG_NUM:02d}_{name}"
    if suffix:
        name = f"{name}_{suffix}"
    if not prefix:
        name = f"{gFIG_NUM:02d}_{name}"
    gFIG_NUM += 1

    # Save the figure:
    plt.savefig(f"{fig_dir}{name}.svg")
    plt.savefig(f"{fig_dir}{name}.png")


def signif(x, p=4):
    """Round the given value(s) in x to the number of significant figures given by p."""
    x_positive = np.where(np.isfinite(x) & (x != 0), np.abs(x), 10**(p-1))
    mags = 10 ** (p - 1 - np.floor(np.log10(x_positive)))
    return np.round(x * mags) / mags


def get_heat_cmap(cmap_name="jet", do_reverse=False):
    """Get the color map of the given name.  Reverse the colors if requested."""
    cmap = plt.get_cmap(cmap_name)
    if do_reverse:
        cmap = cmap.reversed()

    return cmap


def get_zero_white_cmap(cmap_name=None, base_cmap=None):
    """Add white to the bottom of a given colormap so that the minimum value in the map is white."""
    # create colormap
    # Taken from: https://stackoverflow.com/a/49367945
    # ---------------

    # create a colormap that consists of
    # - 1/5 : custom colormap, ranging from white to the first color of the colormap
    # - 4/5 : existing colormap

    # set upper part: 4 * 256/4 entries

    if cmap_name and base_cmap:
        raise RuntimeError("cmap_name and base_cmap are mutually exclusive.")

    if not base_cmap:
        base_cmap = plt.get_cmap(cmap_name)

    upper = base_cmap(np.arange(256))

    # set lower part: 1 * 256/4 entries
    # - initialize all entries to 1 to make sure that the alpha channel (4th column) is 1
    lower = np.ones((int(256 / 4), 4))
    # - modify the first three columns (RGB):
    #   range linearly between white (1,1,1) and the first color of the upper colormap
    for i in range(3):
        lower[:, i] = np.linspace(1, upper[0, i], lower.shape[0])

    # combine parts of colormap
    cmap = np.vstack((lower, upper))

    # convert to matplotlib colormap
    cmap = matplotlib.colors.ListedColormap(cmap, name=f"{cmap_name}ZeroWhite", N=cmap.shape[0])

    return cmap



import pytest
import os
import sys
import subprocess

from click.testing import CliRunner

from longbow.__main__ import main_entry as longbow

from ..utils import assert_reads_files_equal
from ..utils import cat_file_to_pipe
from ..utils import convert_sam_to_bam

################################################################################

TOOL_NAME = "correct"

TEST_DATA_FOLDER = path = os.path.abspath(
    __file__ + os.path.sep + "../../" + os.path.sep + "test_data"
) + os.path.sep + TOOL_NAME + os.path.sep


################################################################################


@pytest.mark.parametrize("input_sam, expected_bc_corrected_sam, expected_bc_uncorrected_sam", [
    [TEST_DATA_FOLDER + "correct_test_data.sam", TEST_DATA_FOLDER + "correct_expected_corrected_data.sam",
     TEST_DATA_FOLDER + "correct_expected_uncorrected_data.sam"],
])
def test_correct(tmpdir, input_sam, expected_bc_corrected_sam, expected_bc_uncorrected_sam):

    # Convert test files to bam:
    input_bam = tmpdir.join("input.bam")
    convert_sam_to_bam(input_sam, input_bam)

    expected_bc_corrected_bam = tmpdir.join("expected.bam")
    convert_sam_to_bam(expected_bc_corrected_sam, expected_bc_corrected_bam)

    actual_bc_corrected_file = tmpdir.join(f"{TOOL_NAME}_actual_out.mas15.bam")
    actual_bc_uncorrected_file = tmpdir.join(f"{TOOL_NAME}_actual_bc_uncorrected_out.mas15.bam")
    args = ["correct", "-t", 1, "-m", "mas_15_sc_10x5p_single_none", "-v", "INFO",
            "-a", f"{TEST_DATA_FOLDER}barcode_allow_list.txt", str(input_bam), "-o", str(actual_bc_corrected_file),
            "--barcode-uncorrectable-bam", str(actual_bc_uncorrected_file)]

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(longbow, args)

    assert result.exit_code == 0

    os.system(f"cp -v {str(actual_bc_corrected_file)} .")

    # Equal files result as True:
    assert_reads_files_equal(actual_bc_corrected_file, expected_bc_corrected_bam, order_matters=True)
    assert_reads_files_equal(actual_bc_uncorrected_file, expected_bc_uncorrected_sam, order_matters=True)


@pytest.mark.skip(reason="`correct` command currently does not accept data from a pipe")
def test_correct_from_pipe(tmpdir, extracted_bam_file_from_pipeline):
    actual_file = tmpdir.join(f"correct_actual_out.pipe.bam")

    proc = subprocess.Popen(
        [sys.executable, "-m", "longbow", "correct", "-f", "-o", actual_file],
        stdin=subprocess.PIPE
    )

    cat_file_to_pipe(extracted_bam_file_from_pipeline, proc)

    assert proc.returncode == 0



import pytest
from click.testing import CliRunner

import subprocess
import sys
import os
import time
import tempfile
import gzip

import pysam

from longbow.__main__ import main_entry as longbow

from ..utils import assert_reads_files_equal
from ..utils import cat_file_to_pipe


TEST_DATA_FOLDER = path = os.path.abspath(
    __file__ + os.path.sep + "../../" + os.path.sep + "test_data"
) + os.path.sep


@pytest.mark.parametrize("input_bam", [
    [TEST_DATA_FOLDER + "mas15_test_input.bam"],
    [TEST_DATA_FOLDER + "mas10_test_input.bam"],
])
def test_convert_from_file(tmpdir, input_bam):
    
    pysam.set_verbosity(0)
    with pysam.AlignmentFile(input_bam[0], "rb", require_index=False, check_sq=False) as input_file:
        with tempfile.NamedTemporaryFile(delete=True, suffix=".fq.gz") as tf:
            with gzip.open(tf.name, "wb") as output_fq:
                for bam_record in input_file:
                    output_fq.write(f'@{bam_record.query_name}\n{bam_record.query_sequence}\n+\n{bam_record.qual}\n'.encode("utf-8"))

            actual_bam = tmpdir.join(f"convert_actual_out.bam")
            args = ["convert", "-f", "-o", actual_bam, input_bam[0]]

            runner = CliRunner()
            result = runner.invoke(longbow, args)

            assert result.exit_code == 0



import pytest
import os
import sys
import subprocess

from click.testing import CliRunner

from longbow.__main__ import main_entry as longbow

from ..utils import cat_file_to_pipe

TEST_DATA_FOLDER = path = os.path.abspath(
    __file__ + os.path.sep + "../../" + os.path.sep + "test_data"
) + os.path.sep


@pytest.mark.parametrize("input_bam, model_exp", [
    [TEST_DATA_FOLDER + "mas15_test_input.bam", "mas_15_sc_10x5p_single_none"],
])
def test_peek_from_file(tmpdir, input_bam, model_exp):

    actual_file = tmpdir.join(f"peek_actual_out.{model_exp}.txt")
    args = ["peek", "-t", 4, input_bam, "-f", "-o", actual_file]

    runner = CliRunner()
    result = runner.invoke(longbow, args)

    assert result.exit_code == 0
    with open(actual_file, 'r') as f:
        l = f.readlines()
        assert l[0].rstrip() == model_exp


@pytest.mark.parametrize("input_bam, model_exp", [
    [TEST_DATA_FOLDER + "mas15_test_input.bam", "mas_15_sc_10x5p_single_none"],
])
def test_peek_from_pipe(tmpdir, input_bam, model_exp):
    actual_file = tmpdir.join(f"peek_actual_out.{model_exp}.pipe.txt")

    proc = subprocess.Popen(
        [ sys.executable, "-m", "longbow", "peek", "-f", "-o", actual_file ],
        stdin=subprocess.PIPE
    )

    cat_file_to_pipe(input_bam, proc)

    assert proc.returncode == 0
    with open(actual_file, 'r') as f:
        l = f.readlines()
        assert l[0].rstrip() == model_exp



import pytest
import os
import sys
import subprocess

from click.testing import CliRunner

from longbow.__main__ import main_entry as longbow

from ..utils import assert_reads_files_equal
from ..utils import cat_file_to_pipe
from ..utils import convert_sam_to_bam

################################################################################

TOOL_NAME = "tagfix"

TEST_DATA_FOLDER = path = os.path.abspath(
    __file__ + os.path.sep + "../../" + os.path.sep + "test_data"
) + os.path.sep + TOOL_NAME + os.path.sep


################################################################################


@pytest.mark.parametrize("input_sam, expected_sam", [
    [
        TEST_DATA_FOLDER + "tagfix_test_data.sam",
        TEST_DATA_FOLDER + "tagfix_test_data.expected.sam"
    ],
])
def test_tagfix(tmpdir, input_sam, expected_sam):

    # Convert test files to bam:
    input_bam = tmpdir.join("input.bam")
    convert_sam_to_bam(input_sam, input_bam)

    expected_bam = tmpdir.join("expected.bam")
    convert_sam_to_bam(expected_sam, expected_bam)

    actual_bam_out = tmpdir.join(f"{TOOL_NAME}_actual_out.bam")
    args = ["tagfix", "-t", 1, "-v", "INFO", "-o", str(actual_bam_out), str(input_bam)]

    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(longbow, args)
    assert result.exit_code == 0

    # Equal files result as True:
    assert_reads_files_equal(actual_bam_out, expected_bam, order_matters=True)


@pytest.mark.skip(reason="tagfix pipe input test is not working properly and needs to be debugged.")
@pytest.mark.parametrize("input_sam, expected_sam", [
    [
        TEST_DATA_FOLDER + "tagfix_test_data.sam",
        TEST_DATA_FOLDER + "tagfix_test_data.expected.sam"
    ],
])
def test_tagfix_from_pipe(tmpdir, input_sam, expected_sam):

    # Convert test files to bam:
    input_bam = tmpdir.join("input.bam")
    convert_sam_to_bam(input_sam, input_bam)

    expected_bam = tmpdir.join("expected.bam")
    convert_sam_to_bam(expected_sam, expected_bam)

    actual_bam_out = tmpdir.join(f"{TOOL_NAME}_actual_out.bam")

    proc = subprocess.Popen(
        [sys.executable, "-m", "longbow", "tagfix", "-t", 1, "-v", "INFO", "-o", str(actual_bam_out)],
        stdin=subprocess.PIPE
    )

    cat_file_to_pipe(str(input_bam), proc)

    assert proc.returncode == 0

    # Equal files result as True:
    assert_reads_files_equal(actual_bam_out, expected_bam, order_matters=True)



import logging
import re
import time
import sys
import itertools

import click
import click_log
import tqdm

import pysam
import multiprocessing as mp

from construct import *

import longbow.utils.constants
from ..utils import bam_utils
from ..utils.cli_utils import format_obnoxious_warning_message


PROG_NAME = "tagfix"

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger(PROG_NAME)
click_log.basic_config(logger)


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-t",
    "--threads",
    type=int,
    default=mp.cpu_count() - 1,
    show_default=True,
    help="number of threads to use (0 for all)",
)
@click.option(
    "-o",
    "--output-bam",
    default="-",
    type=click.Path(exists=False),
    help="annotated bam output  [default: stdout]",
)
@click.option(
    '-f',
    '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help="Force overwrite of the output files if they exist."
)
@click.argument("input-bam", default="-" if not sys.stdin.isatty() else None, type=click.File("rb"))
def main(threads, output_bam, force, input_bam):
    """Update longbow read tags after alignment."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    # Check to see if the output files exist:
    bam_utils.check_for_preexisting_files(output_bam, exist_ok=force)

    threads = mp.cpu_count() if threads <= 0 or threads > mp.cpu_count() else threads
    logger.info(f"Running with {threads} worker subprocess(es)")

    # Configure process manager:
    # NOTE: We're using processes to overcome the Global Interpreter Lock.
    manager = mp.Manager()
    process_input_data_queue = manager.Queue(threads)
    results = manager.Queue()

    pysam.set_verbosity(0)  # silence message about the .bai file not being found
    with pysam.AlignmentFile(
        input_bam, "rb", check_sq=False, require_index=False
    ) as bam_file, tqdm.tqdm(
        desc="Progress",
        unit=" read",
        colour="green",
        file=sys.stderr,
        leave=False,
        disable=not sys.stdin.isatty(),
    ) as pbar:

        # Start worker sub-processes:
        res = manager.dict({"num_tags_corrected": 0, "num_reads": 0})
        worker_process_pool = []
        for _ in range(threads):
            p = mp.Process(
                target=_correct_read_tags, args=(process_input_data_queue, results, bam_file.header, res)
            )
            p.start()
            worker_process_pool.append(p)

        # Check if we have run this command on a bam file already:
        if bam_utils.bam_header_has_longbow_command_program_group(bam_file.header, PROG_NAME):
            logger.error(f"{PROG_NAME} has already been run on input bam: {input_bam}")
            sys.exit(1)

        out_header = bam_utils.create_bam_header_with_program_group(logger.name, bam_file.header)

        # Start output worker:
        output_worker = mp.Process(
            target=_output_writer_fn,
            args=(
                results,
                out_header,
                output_bam,
                pbar,
            ),
        )
        output_worker.start()

        # Add in a sentinel value at the end of the queue - one for each subprocess - so we guarantee
        # that all subprocesses will exit:
        iter_data = itertools.chain(bam_file, (None,) * threads)
        for r in iter_data:
            if r is not None:
                process_input_data_queue.put(r.to_dict())
            else:
                process_input_data_queue.put(r)

        # Wait for our input jobs to finish:
        for p in worker_process_pool:
            p.join()

        # Now that our input processes are done, we can add our exit sentinel onto the output queue and
        # wait for that process to end:
        results.put(None)
        output_worker.join()

    logger.info(f"Corrected tags in {res['num_tags_corrected']} reads of {res['num_reads']} total.")
    
    et = time.time()
    logger.info(f"Done. Elapsed time: {et - t_start:2.2f}s. "
                f"Overall processing rate: {res['num_reads']/(et - t_start):2.2f} reads/s.")

    if res['num_tags_corrected'] == 0:
        logger.warning(format_obnoxious_warning_message("No read tags were corrected.  This is very unlikely.  "
                                                        "You should check your data."))


def _output_writer_fn(out_queue, out_bam_header, out_bam_file_name, pbar):
    """Thread / process fn to write out all our data."""

    with pysam.AlignmentFile(
        out_bam_file_name, "wb", header=out_bam_header
    ) as out_bam_file:
        while True:
            # Wait for some output data:
            raw_data = out_queue.get()

            # Check for exit sentinel:
            if raw_data is None:
                break

            # Unpack data:
            read = raw_data
            read = pysam.AlignedSegment.from_dict(read, out_bam_header)

            # Write out our read:
            out_bam_file.write(read)

            pbar.update(1)


def _correct_read_tags(in_queue, out_queue, bam_header, res):
    """Function to run in each subprocess.
    Do the tag correction here."""

    while True:
        # Wait until we get some data.
        # Note: Because we have a sentinel value None inserted at the end of the input data for each
        #       subprocess, we don't have to add a timeout - we're guaranteed each process will always have
        #       at least one element.
        raw_data = in_queue.get()

        # Check for exit sentinel:
        if raw_data is None:
            return

        # Unpack our data here:
        read = pysam.AlignedSegment.from_dict(raw_data, bam_header)

        if read.is_reverse:

            # Reverse the segments and update the positions:
            segments_tag = read.get_tag(longbow.utils.constants.SEGMENTS_TAG)
            segments = segments_tag.split(longbow.utils.constants.SEGMENT_TAG_DELIMITER)
            read_length = len(read.query_sequence)
            new_segments = []
            for seg in reversed(segments):
                seg_name, start, end = re.split("[:-]", seg)
                new_start = read_length - int(end)
                new_end = read_length - int(start)

                new_segments.append(f"{seg_name}:{new_start}-{new_end}")

            read.set_tag(longbow.utils.constants.SEGMENTS_TAG,
                         longbow.utils.constants.SEGMENT_TAG_DELIMITER.join(new_segments))

            # Reverse the Segment Qualities:
            seg_quals = read.get_tag(longbow.utils.constants.SEGMENTS_QUAL_TAG)
            new_seg_quals = seg_quals.split(longbow.utils.constants.SEGMENT_TAG_DELIMITER)[::-1]
            read.set_tag(longbow.utils.constants.SEGMENTS_QUAL_TAG,
                         longbow.utils.constants.SEGMENT_TAG_DELIMITER.join(new_seg_quals))

            # Increment our counter:
            res["num_tags_corrected"] += 1

        out_queue.put(read.to_dict())
        res["num_reads"] += 1



import json
import logging
import time
import os
import sys

import click
import click_log

import tqdm
import pysam
from construct import *
from collections import Counter

import longbow.utils.constants
from ..utils import bam_utils
from ..utils import model as LongbowModel
from ..utils.model import LibraryModel
from ..annotate.command import get_segments

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("sift")
click_log.basic_config(logger)


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-p",
    "--pbi",
    required=False,
    type=click.Path(exists=True),
    help="BAM .pbi index file",
)
@click.option(
    "-o",
    "--output-bam",
    default="-",
    type=click.Path(exists=False),
    help="filtered bam output (passing reads only)  [default: stdout]",
)
@click.option(
    "-x",
    "--reject-bam",
    default="/dev/null",
    type=click.Path(exists=False),
    help="Filtered bam output (failing reads only)  [default: /dev/null]",
)
@click.option(
    "-m",
    "--model",
    help="The model to use for annotation.  If not specified, it will be autodetected from "
         "the BAM header.  If the given value is a pre-configured model name, then that "
         "model will be used.  Otherwise, the given value will be treated as a file name "
         "and Longbow will attempt to read in the file and create a LibraryModel from it.  "
         "Longbow will assume the contents are the configuration of a LibraryModel as per "
         "LibraryModel.to_json()."
)
@click.option(
    "-l",
    "--validation-model",
    default="10x_sc_10x5p_single_none",
    help="The model to use for cDNA validation."
)
@click.option(
    '-f',
    '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help="Force overwrite of the output files if they exist."
)
@click.option(
    "-s",
    "--stats",
    default="/dev/null",
    type=click.Path(exists=False),
    help="Table describing the ways in which the reads do not conform to expectation (failing reads only)"
         "[default: /dev/null]",
)
@click.option(
    "-u",
    "--summary-stats",
    default="/dev/null",
    type=click.Path(exists=False),
    help="Table containing summary statistics for the sifted reads that are output."
         "[default: /dev/null]",
)
@click.option(
    "-k",
    "--ignore-list",
    required=False,
    type=click.Path(exists=True),
    help="Txt file containing a list of read names to ignore.",
)
@click.argument("input-bam", default="-" if not sys.stdin.isatty() else None, type=click.File("rb"))
def main(pbi, output_bam, reject_bam, model, validation_model, force, stats, summary_stats, ignore_list, input_bam):
    """Filter segmented reads by conformation to expected cDNA design."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    pbi = f"{input_bam.name}.pbi" if pbi is None else pbi
    read_count = None
    if os.path.exists(pbi):
        read_count = bam_utils.load_read_count(pbi)
        logger.info("About to Filter %d reads", read_count)

    reads_to_ignore = set()
    if ignore_list and os.path.exists(ignore_list):
        logger.info(f"Ingesting read ignore list: {ignore_list}")
        with open(ignore_list, 'r') as f:
            for line in f:
                read_name = line.strip()
                if len(read_name) > 0:
                    reads_to_ignore.add(read_name)
        logger.info(f"Num reads to ignore: {len(reads_to_ignore)}")

    # Check to see if the output files exist:
    bam_utils.check_for_preexisting_files([output_bam, reject_bam], exist_ok=force)

    logger.info(f"Writing reads that conform to the model to: {output_bam}")
    logger.info(f"Writing reads that do not conform to the model to: {reject_bam}")

    # Open our input bam file:
    pysam.set_verbosity(0)
    with pysam.AlignmentFile(input_bam, "rb", check_sq=False, require_index=False) as bam_file, \
            tqdm.tqdm(
            desc="Progress",
            unit=" read",
            colour="green",
            file=sys.stderr,
            disable=not sys.stdin.isatty(),
            total=read_count
            ) as pbar:

        # Get our model:
        if model is None:
            lb_model = LibraryModel.from_json_obj(bam_utils.get_model_from_bam_header(bam_file.header))
        elif model is not None and LibraryModel.has_prebuilt_model(model):
            lb_model = LibraryModel.build_pre_configured_model(model)
        else:
            lb_model = LibraryModel.from_json_file(model)

        # TODO: Currently only one model works.  FIX THIS.
        # Make sure we've specified the model that works:
        if validation_model != "10x_sc_10x5p_single_none":
            logger.error(f"Given validation model is not yet implemented with `longbow sift`: {validation_model}")
            sys.exit(1)

        if lb_model.name != "mas_15_sc_10x5p_single_none":
            logger.error(f"Given model is not yet implemented with `longbow sift`: {lb_model.name}")
            sys.exit(1)

        logger.info(f"Using %s: %s", lb_model.name, lb_model.description)

        # Get our header from the input bam file:
        out_header = bam_utils.create_bam_header_with_program_group(logger.name, bam_file.header, models=[lb_model])

        # Setup output files:
        with pysam.AlignmentFile(output_bam, "wb", header=out_header) as passing_bam_file, \
                pysam.AlignmentFile(reject_bam, "wb", header=out_header) as failing_bam_file, \
                open(stats, 'w') as stats_file:

            num_passed = 0
            num_failed = 0
            num_ignored = 0

            # Create the forward-connected model of just this section of the read
            sub_model = LibraryModel.build_pre_configured_model(validation_model)
            sub_model.build_forward_connected()

            stats_file.write('\t'.join(['read_name', 'rq', '5p_Adapter', 'CBC', 'UMI', 'SLS', 'cDNA', 'Poly_A', '3p_Adapter', 'SG']) + '\n')

            for read in bam_file:

                if read.query_name in reads_to_ignore:
                    logger.debug(f"Ignoring read: {read.query_name}")
                    num_ignored += 1
                    continue

                # Get our read segments:
                try:
                    _, segments = get_segments(read)
                except KeyError:
                    logger.error(f"Input bam file does not contain longbow segmented reads!  "
                                 f"No {longbow.utils.constants.SEGMENTS_TAG} tag detected on read {read.query_name} !")
                    sys.exit(1)

                # Annotate the read with the model that was used in its validation:
                logp, ppath = sub_model.annotate(read.query_sequence)
                qpath = bam_utils.collapse_annotations(ppath)

                counts = Counter([q.to_tag().split(":")[0] for q in qpath])

                is_valid = (counts['5p_Adapter'] == 1 and
                            counts['CBC'] == 1 and
                            counts['UMI'] == 1 and
                            counts['SLS'] == 1 and
                            counts['cDNA'] == 1 and
                            counts['Poly_A'] == 1 and
                            counts['3p_Adapter'] == 1)

                read.set_tag(longbow.utils.constants.READ_MODEL_NAME_TAG, lb_model.name)
                read.set_tag(longbow.utils.constants.SEGMENTS_TAG, longbow.utils.constants.SEGMENT_TAG_DELIMITER.join([q.to_tag() for q in qpath]))
                read.set_tag(longbow.utils.constants.READ_IS_VALID_FOR_MODEL_TAG, is_valid)

                if is_valid:
                    logger.debug("Read is %s valid: %s: adapter pattern: %s",
                                 lb_model.name,
                                 read.query_name,
                                 read.get_tag(longbow.utils.constants.SEGMENTS_TAG))

                    passing_bam_file.write(read)
                    num_passed += 1
                else:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Read is not %s valid: %s: adapter pattern: %s",
                                     lb_model.name,
                                     read.query_name,
                                     read.get_tag(longbow.utils.constants.SEGMENTS_TAG))

                    failing_bam_file.write(read)
                    num_failed += 1

                    stats_file.write('\t'.join([
                        read.query_name,
                        str(read.get_tag("rq")),
                        str(counts['5p_Adapter']),
                        str(counts['CBC']),
                        str(counts['UMI']),
                        str(counts['SLS']),
                        str(counts['cDNA']),
                        str(counts['Poly_A']),
                        str(counts['3p_Adapter']),
                        longbow.utils.constants.SEGMENT_TAG_DELIMITER.join([q.to_tag() for q in qpath])
                        ]) + '\n'
                    )

                pbar.update(1)

    # Calc some stats:
    tot_reads = num_passed + num_failed + num_ignored
    pct_reads_passing = 100 * num_passed / tot_reads if tot_reads > 0 else 0
    pct_reads_failing = 100 * num_failed / tot_reads if tot_reads > 0 else 0
    pct_reads_ignored = 100 * num_ignored / tot_reads if tot_reads > 0 else 0

    # Yell at the user / write out summary stats:
    with open(summary_stats, 'w') as f:
        message = f"Total Reads Processed:\t{tot_reads:d}"
        f.write(f"{message}\n")
        logger.info(message)

        message = f"# Reads Passing Model Filter:\t{num_passed:d}\t{pct_reads_passing:02.4f}"
        f.write(f"{message}\n")
        logger.info(message)

        message = f"# Reads Failing Model Filter:\t{num_failed:d}\t{pct_reads_failing:02.4f}"
        f.write(f"{message}\n")
        logger.info(message)

        message = f"# Reads Ignored:\t{num_ignored:d}\t{pct_reads_ignored:02.4f}"
        f.write(f"{message}\n")
        logger.info(message)

    logger.info(f"Done. Elapsed time: %2.2fs.", time.time() - t_start)



import pytest

import subprocess
import sys
import os

from click.testing import CliRunner

from longbow.__main__ import main_entry as longbow

from ..utils import assert_reads_files_equal
from ..utils import cat_file_to_pipe

################################################################################

TOOL_NAME = "annotate"

TEST_DATA_FOLDER = path = os.path.abspath(
    __file__ + os.path.sep + "../../" + os.path.sep + "test_data"
) + os.path.sep
EXPECTED_DATA_FOLDER = TEST_DATA_FOLDER + TOOL_NAME + os.path.sep

################################################################################


@pytest.mark.parametrize("input_bam, expected_bam, model_name", [
    [TEST_DATA_FOLDER + "mas15_test_input.bam", EXPECTED_DATA_FOLDER + "mas15v2_expected.bam", "mas_15_sc_10x5p_single_none"],
    [TEST_DATA_FOLDER + "mas10_test_input.bam", EXPECTED_DATA_FOLDER + "mas10v2_expected.bam", "mas_10_sc_10x5p_single_none"],
])
def test_annotate(tmpdir, input_bam, expected_bam, model_name):

    actual_file = tmpdir.join(f"{TOOL_NAME}_actual_out.{model_name}.bam")
    args = ["annotate", "-t", 1, "-v", "INFO", "-m", model_name, input_bam, "-o", str(actual_file)]

    runner = CliRunner()
    result = runner.invoke(longbow, args)

    os.system(f"cp {str(actual_file)} .")

    assert result.exit_code == 0
    assert_reads_files_equal(actual_file, expected_bam, order_matters=True)


@pytest.mark.parametrize("input_bam, expected_bam, model_name", [
    [TEST_DATA_FOLDER + "mas15_test_input.bam", EXPECTED_DATA_FOLDER + "mas15v2_expected.bam", "mas_15_sc_10x5p_single_none"],
    [TEST_DATA_FOLDER + "mas10_test_input.bam", EXPECTED_DATA_FOLDER + "mas10v2_expected.bam", "mas_10_sc_10x5p_single_none"],
])
def test_annotate_from_pipe(tmpdir, input_bam, expected_bam, model_name):
    actual_bam = tmpdir.join(f"annotate_actual_out.{model_name}.pipe.bam")

    proc = subprocess.Popen(
        [sys.executable, "-m", "longbow", "annotate", "-m", model_name, "-f", "-o", actual_bam],
        stdin=subprocess.PIPE
    )

    cat_file_to_pipe(input_bam, proc)

    assert proc.returncode == 0
    assert_reads_files_equal(actual_bam, expected_bam)
